from typing import Any

from dliswriter.logical_record.core.eflr import EFLRSet, EFLRItem
from dliswriter.utils.internal.internal_enums import EFLRType
from dliswriter.logical_record.core.attribute import TextAttribute


class LongNameItem(EFLRItem):
    """Model an object being part of LongName EFLR."""

    parent: "LongNameSet"

    def __init__(self, name: str, parent: "LongNameSet", **kwargs: Any) -> None:
        """Initialise LongNameItem.

        Args:
            name        :   Name of the LongNameItem.
            parent      :   Parent LongNameSet of this LongNameItem.
            **kwargs    :   Values of to be set as characteristics of the LongNameItem Attributes.
        """

        self.general_modifier = TextAttribute('general_modifier', multivalued=True)
        self.quantity = TextAttribute('quantity')
        self.quantity_modifier = TextAttribute('quantity_modifier', multivalued=True)
        self.altered_form = TextAttribute('altered_form')
        self.entity = TextAttribute('entity')
        self.entity_modifier = TextAttribute('entity_modifier', multivalued=True)
        self.entity_number = TextAttribute('entity_number')
        self.entity_part = TextAttribute('entity_part')
        self.entity_part_number = TextAttribute('entity_part_number')
        self.generic_source = TextAttribute('generic_source')
        self.source_part = TextAttribute('source_part', multivalued=True)
        self.source_part_number = TextAttribute('source_part_number', multivalued=True)
        self.conditions = TextAttribute('conditions', multivalued=True)
        self.standard_symbol = TextAttribute('standard_symbol')
        self.private_symbol = TextAttribute('private_symbol')

        super().__init__(name, parent=parent, **kwargs)


class LongNameSet(EFLRSet):
    """Model LongName EFLR."""

    set_type = 'LONG-NAME'
    logical_record_type = EFLRType.LNAME
    item_type = LongNameItem


LongNameItem.parent_eflr_class = LongNameSet



import pytest

from dliswriter.logical_record.eflr_types.file_header import FileHeaderSet, FileHeaderItem
from dliswriter import high_compatibility_mode_decorator


@pytest.mark.parametrize(("header_id", "identifier", "sequence_number"), (("a", "A", 9), ("123ert", "8", 3)))
def test_creation(header_id: str, identifier: str, sequence_number: int) -> None:
    """Test creating a FileHeaderItem."""

    fh = FileHeaderItem(header_id, sequence_number=sequence_number, parent=FileHeaderSet())

    assert fh.header_id == header_id
    assert fh.sequence_number == sequence_number

    assert isinstance(fh.parent, FileHeaderSet)
    assert fh.parent.set_name is None


@pytest.mark.parametrize("name", ("FH-1", "DEFAULT_FILE_HEADER", "12-3"))
@high_compatibility_mode_decorator
def test_header_id_compatible(name: str) -> None:
    FileHeaderItem(name, parent=FileHeaderSet())


@pytest.mark.parametrize("name", ("Fh1", "file.header", "fh 8"))
@high_compatibility_mode_decorator
def test_header_id_not_compatible(name: str) -> None:
    with pytest.raises(ValueError, match=".*strings can contain only uppercase characters, digits, dashes, .*"):
        FileHeaderItem(name, parent=FileHeaderSet())



from enum import Enum, IntEnum
from struct import Struct
from typing import Union, Any
from typing_extensions import Self


class RepresentationCode(int, Enum):
    """Collect RP66 V1 representation codes with Struct converters where possible.

    Some representation codes can not be directly converted using Struct.
    In these cases, the converter is set to None. These codes are still included here,
    so that the enumeration also serves as a full list of representation codes included in the RP66 V1.

    The overridden '__new__' method assigns a new 'converter' attribute to each enum member. This facilitates calling
    the member's converter on any value, e.g.:
        RepresentationCode.FDOUBL.converter.pack(<value>)
    """

    converter: Union[Struct, None]

    def __new__(cls, code: int, converter: Union[Struct, None] = None) -> "RepresentationCode":
        """When a new member is created, assign not only the integer value, but also a converter.

        Args:
            code        :   Integer value of the enum member.
            converter   :   Struct which can be used to convert a value to bytes according to the format specified
                            when the Struct is initialised.
        """

        obj = super().__new__(cls, code)
        obj._value_ = code
        obj.converter = converter
        return obj

    FSHORT = 1, Struct('>h')
    FSINGL = 2, Struct('>f')
    FSING1 = 3, Struct('>ff')
    FSING2 = 4, Struct('>fff')
    ISINGL = 5, Struct('>i')
    VSINGL = 6, Struct('>i')
    FDOUBL = 7, Struct('>d')
    FDOUB1 = 8, Struct('>dd')
    FDOUB2 = 9, Struct('>ddd')
    CSINGL = 10, Struct('>ff')
    CDOUBL = 11, Struct('>dd')
    SSHORT = 12, Struct('>b')
    SNORM = 13, Struct('>h')
    SLONG = 14, Struct('>i')
    USHORT = 15, Struct('>B')
    UNORM = 16, Struct('>H')
    ULONG = 17, Struct('>I')
    UVARI = 18, None
    IDENT = 19, None
    ASCII = 20, None
    DTIME = 21, Struct('>BBBBBBH')
    ORIGIN = 22, None
    OBNAME = 23, None
    OBJREF = 24, None
    ATTREF = 25, None
    STATUS = 26, Struct('>B')

    def convert(self, value: Any) -> bytes:
        """Convert the given value to bytes according to the Struct pattern defined for the enum member (if any)."""

        if self.converter is None:
            raise RuntimeError("Converter struct not defined; cannot directly convert the value to bytes")
        return self.converter.pack(value)

    def decode_bytes(self, value: bytes) -> tuple[Any, ...]:
        """Convert bytes to a value according to the Struct pattern defined for the enum member (if any).

        Used in unit tests.
        """

        if self.converter is None:
            raise RuntimeError("Converter struct not defined; cannot directly decode the bytes")

        s = self.converter.size
        v = len(value)
        if v % s:
            raise ValueError(f"Size of the provided bytes must be an integer multiple of {s}; got {v}")

        return tuple(self.converter.unpack(value[i*s:(i+1)*s])[0] for i in range(v//s))

    @classmethod
    def get_member(cls, v: Union[str, int, None, Self], allow_none: bool = False) -> Union[Self, None]:
        """Helper function: get a member of the RepresentationCode enum, given the name, value, or the member itself.

        Args:
            v           :   Name or value of the enum, the enum member itself, or None (see below).
            allow_none  :   If True and v is None, return None. Otherwise, a ValueError will be raised.

        Returns:
            The member of the enumeration corresponding to the provided value/name/member (or None).
        """

        if allow_none and v is None:
            return None

        if isinstance(v, cls):
            return v

        if isinstance(v, int):
            try:
                return cls(v)
            except ValueError:
                pass

        if isinstance(v, str):
            try:
                return cls[v]
            except KeyError:
                pass

        raise ValueError(f"{cls.__name__} '{v}' is not defined")


class EFLRType(IntEnum):
    """Types of explicitly formatted logical records."""

    FHLR = 0
    OLR = 1
    AXIS = 2
    CHANNL = 3
    FRAME = 4
    STATIC = 5
    SCRIPT = 6
    UPDATE = 7
    UDI = 8
    LNAME = 9
    SPEC = 10
    DICT = 11


class IFLRType(IntEnum):
    """Types of indirectly formatted logical records."""

    FDATA = 0
    NOFMT = 1



import logging
from functools import cached_property
from typing import TYPE_CHECKING, Any, Union, Optional, Generator
import numpy as np

from dliswriter.utils.internal.struct_writer import write_struct_obname
from dliswriter.logical_record.core.attribute.attribute import Attribute
from dliswriter.utils.internal.value_checkers import validate_string
from dliswriter.utils import enums

if TYPE_CHECKING:
    from dliswriter.logical_record.core.eflr.eflr_set import EFLRSet
    from dliswriter.logical_record.core.attribute.subtypes import DimensionAttribute


logger = logging.getLogger(__name__)


class AttrSetup:
    """Convenience class to pass setup for an Attribute without creating a dictionary."""

    # Note: the initial plan was to make it a dataclass, but this doesn't give the PyCharm IDE typing hints,
    # which is one of the main points of this class

    def __init__(self, value: Optional[Any] = None, units: Optional[Union[str, enums.Unit]] = None) -> None:
        """Initialise AttrSetup with any of the passed values."""

        self.value = value
        self.units = units
        # Note: no checks done here, because they're done in EFLRItem/Attribute later anyway

    def items(self) -> Generator:
        for item_name in ('value', 'units'):
            if (item_value := getattr(self, item_name)) is not None:
                yield item_name, item_value


class EFLRItem:
    """Model an item belonging to an Explicitly Formatted Logical Record - e.g. a particular channel."""

    parent_eflr_class: type["EFLRSet"] = NotImplemented

    def __init__(self, name: str, parent: "EFLRSet", origin_reference: Optional[int] = None, **kwargs: Any) -> None:
        """Initialise an EFLRItem.

        Args:
            name                :   Name of the item. This will be the name it is stored with in the created DLIS file.
            parent              :   EFLRTable instance this item belongs to. If not provided,
                                        retrieved/made based on set_name.
            origin_reference    :   Common for EFLRItems sharing the same Origin.
            **kwargs            :   Values to be set in attributes of this item.

        Note:
            When a subclass of EFLRItem is defined, all the attributes should be defined before calling
            super().__init__. This makes it possible for values of attributes to be set here, through 'set_attributes'
            method call.

        """

        self.name = validate_string(name)    #: name of the item

        self._check_parent(parent)
        self._parent = parent  #: EFLRSet instance this item belongs to
        self._parent.register_item(self)

        #: origin reference value, common for records sharing origin
        self._origin_reference: Union[int, None] = self._validate_origin_reference(origin_reference, allow_none=True)

        #: copy number of the item - ith EFLRItem of the same name and type
        self._copy_number = self._compute_copy_number()

        for attribute in self.attributes.values():
            attribute.parent_eflr = self

        self.set_attributes(**{k: v for k, v in kwargs.items() if v is not None})

    @property
    def parent(self) -> "EFLRSet":
        """EFLRSet instance that this EFLRItem is a part of."""

        return self._parent

    @property
    def copy_number(self) -> int:
        """Copy number of this EFLRItem."""

        return self._copy_number

    @property
    def origin_reference(self) -> Union[int, None]:
        """Reference to the Origin associated with this EFLRItem."""

        return self._origin_reference

    @origin_reference.setter
    def origin_reference(self, v: int) -> None:
        """Set a new origin reference (point to a different Origin)."""

        self._origin_reference = self._validate_origin_reference(v)

    @staticmethod
    def _validate_origin_reference(v: Union[int, None], allow_none: bool = False) -> Union[int, None]:
        """Check that the provided origin reference value is an integer (or None, depending on the setup).

        Return the provided origin reference value (if it's correct). Raise an error otherwise.
        """

        if v is None and allow_none:
            return None

        if not isinstance(v, int):
            raise TypeError(f"Origin reference must be an integer; got {type(v)}: {v}")

        return v

    def _compute_copy_number(self) -> int:
        """Compute copy number of this ELFRItem, i.e. how many other objects of the same type and name there are."""

        items_with_the_same_name = filter(lambda o: o.name == self.name, self.parent.get_all_eflr_items())
        return len(list(items_with_the_same_name)) - 1

    @classmethod
    def _check_parent(cls, parent: "EFLRSet") -> None:
        """Validate a parent EFLRSet instance for this EFLRItem.

        Args:
            parent      :   Parent EFLRSet instance.

        Returns:
            The parent EFLRSet instance.
        """

        if not isinstance(parent, cls.parent_eflr_class):
            raise TypeError(f"Expected an instance of {cls.parent_eflr_class.__name__}; "
                            f"got a {type(parent)}: {parent}")

    @property
    def attributes(self) -> dict[str, Attribute]:
        """Attributes defined for this EFLRItem (sub)class with its values for the current instance."""

        return {key: value for key, value in self.__dict__.items() if isinstance(value, Attribute)}

    def __str__(self) -> str:
        """Description of the EFLRItem instance."""

        return f"{self.__class__.__name__} '{self.name}'"

    def __setattr__(self, key: str, value: Any) -> None:
        """Limit the possibility of setting attributes by excluding Attribute instances.

        This prevents overwriting Attribute instances being attributes of this EFLRItem. ValueError is raised at such
        attempt. If only a value of the Attribute instance is supposed to be changed, the Attribute's 'value' attribute
        should be used instead.
        """

        if isinstance(getattr(self, key, None), Attribute):
            raise RuntimeError(f"Cannot set DLIS Attribute '{key}'. Did you mean setting '{key}.value' instead?")

        return super().__setattr__(key, value)

    @cached_property
    def obname(self) -> bytes:
        """Create OBNAME bytes of this item - bytes used to identify an item in the file.

        They serve as a reference to the current item - e.g. when a Parameter references a Zone.
        """

        return write_struct_obname(self)

    def _make_attrs_bytes(self) -> bytes:
        """Create bytes describing the values of the EFLRItem instance's Attributes."""

        _bytes = b''
        for attr in self.attributes.values():
            if attr.value is None:
                _bytes += b'\x00'
            else:
                _bytes += attr.get_as_bytes()

        return _bytes

    def _run_checks_and_set_defaults(self) -> None:
        """Called before writing the item's bytes. Set default values to some attributes if they were not set at all."""

        pass

    def make_item_body_bytes(self) -> bytes:
        """Create bytes describing the item: its name and values of its attributes."""

        self._run_checks_and_set_defaults()

        return b'p' + self.obname + self._make_attrs_bytes()

    def set_attributes(self, **kwargs: Any) -> None:
        """Set the values and other characteristics of the EFLRItem's attributes.

        Args:
            **kwargs    :   The mapping of attribute names (and parts) on their values to be set.

        This method allows setting 'value', 'units', and 'representation_code' parts of any attribute of
        the item. The expected syntax for the arguments' keys is <attr_name>.<part_name>, e.g.: 'dimension.value',
        'minimum_value.representation_code', 'minimum_value.units'. It is also possible to use only the attribute name
        as the key (e.g. 'dimension', 'maximum_value'); in this case, the 'value' part of the attribute (dimension.value
        or maximum_value.value respectively) is set to the value of the keyword argument.
        """

        def set_value(_attr: Attribute, _value: Any, _key: str = 'value') -> None:
            """Set 'value' or 'units' of the provided Attribute instance."""

            logger.debug(f"Setting {_attr.label}.{_key} of {self} to {repr(_value)}")
            setattr(_attr, _key, _value)

        for attr_name, attr_value in kwargs.items():
            attr = getattr(self, attr_name, None)
            if not attr or not isinstance(attr, Attribute):
                raise AttributeError(f"{self.__class__.__name__} does not have attribute '{attr_name}'")

            if isinstance(attr_value, (dict, AttrSetup)):
                for key, value in attr_value.items():
                    if key not in ('value', 'units'):
                        raise ValueError(f"Cannot set {key} of a(n) {attr.__class__.__name__}")
                    set_value(attr, value, key)

            else:
                set_value(attr, attr_value)

    @staticmethod
    def count_attributes(*attrs: Attribute) -> dict:
        """Count the number of values (length of value list) defined for each of the provided Attribute.

        If no values are defined foe an Attribute, it is not included in the counts dict.
        """

        value_counts = {}
        for attr in attrs:
            if attr.value is not None:
                value_counts[attr.label] = len(attr.value)

        return value_counts


class DimensionedItem:
    """Mixin to be used with EFLRItem subclasses which define 'axis' and 'dimension' Attributes."""

    axis: Attribute
    dimension: "DimensionAttribute"

    def _check_axis_vs_dimension(self) -> None:
        """Check that the number of axes matches the number of dimensions defined for the EFLRItem."""

        axs = self.axis.value
        dims = self.dimension.value

        if axs is None:
            return
        if dims is None:
            return

        if (na := len(axs)) != (nd := len(dims)):
            raise RuntimeError(f"{self}: number of axes ({na}) does not match the number of dimensions ({nd})")

        for i in range(na):
            ac = axs[i].coordinates.value
            if ac is None:
                continue
            if (nc := len(ac)) != dims[i]:
                raise RuntimeError(f"{self}: number of coordinates in axis {i+1} ({nc}) does not match the "
                                   f"dimension {i+1} ({dims[i]})")

    def _check_or_set_value_dimensionality(self, value: Union[list, tuple, None],
                                           value_label: Optional[str] = None) -> None:
        """Determine the dimensionality (shape) of a value. Verify or set up the 'dimension' attr based on that."""

        if value is None:
            return

        value_label = value_label or 'value'

        try:
            arr = np.array(value)
        except ValueError:
            raise RuntimeError(f"{self}: {value_label} {value} does not have a regular dimensionality structure")

        dim_from_value = list(arr.shape[1:])
        if self.dimension.value is not None:
            if dim_from_value != self.dimension.value:
                raise RuntimeError(f"{self}: shape of {value_label} {value} (shape {arr.shape}) does not match "
                                   f"the specified dimensionality: {self.dimension.value}")
        else:
            self.dimension.value = dim_from_value



import pytest
from typing import Any

from dliswriter.logical_record.eflr_types.parameter import ParameterSet, ParameterItem
from dliswriter.utils.internal.internal_enums import RepresentationCode


@pytest.mark.parametrize(("name", "value", "value_type", "n_values", "repr_code"), (
        ("p1", ["ahaha", "abc"], str, 2, RepresentationCode.ASCII),
        ("some param", [2.33, 2.1], float, 2, RepresentationCode.FDOUBL),
        ("p89", -12.1211, float, 1, RepresentationCode.FDOUBL)
))
def test_from_config(name: str, value: Any, value_type: type, n_values: int, repr_code: RepresentationCode) -> None:
    """Test creating ParameterItem."""

    param = ParameterItem(name, values=value, parent=ParameterSet())

    assert param.name == name

    assert param.values.representation_code is repr_code
    assert isinstance(param.values.value, list)
    assert all(isinstance(v, value_type) for v in param.values.value)
    assert len(param.values.value) == n_values

    assert isinstance(param.parent, ParameterSet)
    assert param.parent.set_name is None



import logging
from typing import Any

from dliswriter.logical_record.core.eflr import EFLRSet, EFLRItem
from dliswriter.utils.internal.internal_enums import EFLRType
from dliswriter.logical_record.eflr_types.channel import ChannelSet
from dliswriter.logical_record.eflr_types.parameter import ParameterSet
from dliswriter.logical_record.eflr_types.calibration_measurement import CalibrationMeasurementSet
from dliswriter.logical_record.eflr_types.calibration_coefficient import CalibrationCoefficientSet
from dliswriter.logical_record.core.attribute import EFLRAttribute, IdentAttribute


logger = logging.getLogger(__name__)


class CalibrationItem(EFLRItem):
    """Model an object being part of Calibration EFLR."""

    parent: "CalibrationSet"

    def __init__(self, name: str, parent: "CalibrationSet", **kwargs: Any) -> None:
        """Initialise CalibrationItem.

        Args:
            name        :   Name of the CalibrationItem.
            parent      :   Parent CalibrationSet of this CalibrationItem.
            **kwargs    :   Values of to be set as characteristics of the CalibrationItem Attributes.
        """

        self.calibrated_channels = EFLRAttribute(
            'calibrated_channels', object_class=ChannelSet, multivalued=True)
        self.uncalibrated_channels = EFLRAttribute(
            'uncalibrated_channels', object_class=ChannelSet, multivalued=True)
        self.coefficients = EFLRAttribute(
            'coefficients', object_class=CalibrationCoefficientSet, multivalued=True)
        self.measurements = EFLRAttribute(
            'measurements', object_class=CalibrationMeasurementSet, multivalued=True)
        self.parameters = EFLRAttribute('parameters', object_class=ParameterSet, multivalued=True)
        self.method = IdentAttribute('method')

        super().__init__(name, parent=parent, **kwargs)


class CalibrationSet(EFLRSet):
    """Model Calibration EFLR."""

    set_type = 'CALIBRATION'
    logical_record_type = EFLRType.STATIC
    item_type = CalibrationItem


CalibrationItem.parent_eflr_class = CalibrationSet



import logging
from typing import Any

from dliswriter.logical_record.core.eflr import EFLRSet, EFLRItem
from dliswriter.logical_record.eflr_types.channel import ChannelSet
from dliswriter.logical_record.eflr_types.computation import ComputationSet
from dliswriter.logical_record.eflr_types.parameter import ParameterSet
from dliswriter.utils.internal.internal_enums import EFLRType
from dliswriter.utils.enums import ProcessStatus
from dliswriter.logical_record.core.attribute import EFLRAttribute, TextAttribute, IdentAttribute, PropertiesAttribute


logger = logging.getLogger(__name__)


class ProcessItem(EFLRItem):
    """Model an object being part of Process EFLR."""

    parent: "ProcessSet"

    def __init__(self, name: str, parent: "ProcessSet", **kwargs: Any) -> None:
        """Initialise ProcessItem.

        Args:
            name        :   Name of the ProcessItem.
            parent      :   Parent ProcessSet of this ProcessItem.
            **kwargs    :   Values of to be set as characteristics of the ProcessItem Attributes.
        """

        self.description = TextAttribute('description')
        self.trademark_name = TextAttribute('trademark_name')
        self.version = TextAttribute('version')
        self.properties = PropertiesAttribute('properties')
        self.status = IdentAttribute(
            'status', converter=ProcessStatus.make_converter("status values"))
        self.input_channels = EFLRAttribute('input_channels', object_class=ChannelSet, multivalued=True)
        self.output_channels = EFLRAttribute('output_channels', object_class=ChannelSet, multivalued=True)
        self.input_computations = EFLRAttribute('input_computations', object_class=ComputationSet, multivalued=True)
        self.output_computations = EFLRAttribute('output_computations', object_class=ComputationSet, multivalued=True)
        self.parameters = EFLRAttribute('parameters', object_class=ParameterSet, multivalued=True)
        self.comments = TextAttribute('comments', multivalued=True)

        super().__init__(name, parent=parent, **kwargs)


class ProcessSet(EFLRSet):
    """Model Process EFLR."""

    set_type = 'PROCESS'
    logical_record_type = EFLRType.STATIC
    item_type = ProcessItem


ProcessItem.parent_eflr_class = ProcessSet



"""Create synthetic data with a given number of rows and optionally, given number and width of images (2D data).
Define channels for the data and write them into a DLIS file.
"""

import logging
from argparse import ArgumentParser
from pathlib import Path
import numpy as np

from dliswriter import DLISFile, enums

from utils import prepare_directory, install_colored_logger


logger = logging.getLogger(__name__)


def make_parser() -> ArgumentParser:
    """Define an argument parser for defining the DLIS file to be created."""

    parser = ArgumentParser("DLIS file creation")
    parser.add_argument('-ofn', '--output-file-name', help='Output file name', required=True)
    parser.add_argument('--time-based', action='store_true', default=False,
                        help="Make a time-based DLIS file (default is depth-based)")
    parser.add_argument('--input-chunk-size', default=1e5, type=float,
                        help="Chunk size (number of rows) for the input data to be processed in")
    parser.add_argument('--output-chunk-size', default=2 ** 32, type=float,
                        help="Size (in bytes) of the chunks in which the output file will be created")
    parser.add_argument('--overwrite', action='store_true', default=False,
                        help="Allow overwriting existing output file")
    parser.add_argument('-vrl', '--visible-record-length', type=int, default=8192,
                        help="Maximum allowed visible record length")

    parser.add_argument('-n', '--n-points', help='Number of data points', type=float, default=10e3)
    parser.add_argument('-ni', '--n-images', type=int, default=0,
                        help='Number of 2D data sets to add (ignored if input file is specified)')
    parser.add_argument('-nc', '--n-cols', type=int, default=128,
                        help='No. columns for each of the added 2D data sets (ignored if input file specified)')

    return parser


def make_image(n_points: int, n_cols: int, divider: int = 11) -> np.ndarray:
    """Create a 2D array with synthetic data.

    Args:
        n_points    :   Number of rows.
        n_cols      :   Number of columns.
        divider     :   The array will be populated with remainders of division of values in range
                        (0, n_points x n_cols) by the divider. This creates an array with periodically repeating
                        integers, facilitating fast visual checks whether file data has been saved and read correctly.

    Returns:
        A numpy.ndarray of shape (n_points, n_cols) with values in range (0, divider).
    """

    return (np.arange(n_points * n_cols) % divider).reshape(n_points, n_cols) + 5 * np.random.rand(n_points, n_cols)


def create_dlis_spec(n_points: int, n_images: int = 0, n_cols: int = 128, time_based: bool = False) -> DLISFile:
    """Create a DLIS file specification (DLISFile object) according to the arguments setup.

    This created 3 scalar (1D) datasets ('time' or 'depth', 'rpm', and 'col3') and as many 2D datasets as specified
    by the provided arguments.

    Args:
        n_points    :   Length (number of points/rows) for all datasets.
        n_images    :   Number of 2D datasets to be added.
        n_cols      :   Number of columns for each 2D dataset.
        time_based  :   If True, the first created dataset will be 'time'. Otherwise, it will be 'depth'.
    """

    df = DLISFile()
    lf = df.add_logical_file()
    lf.add_origin("ORIGIN")

    if time_based:
        logger.debug("Creating time dataset")
        index = lf.add_channel("TIME", data=0.5 * np.arange(n_points))
        index_type = enums.FrameIndexType.NON_STANDARD
    else:
        logger.debug("Creating depth dataset")
        index = lf.add_channel('DEPTH', data=2500 + 0.1 * np.arange(n_points))
        index_type = enums.FrameIndexType.BOREHOLE_DEPTH

    channels = [index]

    for i in range(n_images):
        logger.debug(f"Creating image dataset {i+1}/{n_images}")
        ch = lf.add_channel(
            f'IMAGE{i}', data=make_image(n_points, n_cols, divider=int(10 + (n_cols - 11) * np.random.rand())))
        channels.append(ch)

    lf.add_frame("MAIN-FRAME", channels=channels, index_type=index_type)

    return df


def main() -> None:
    install_colored_logger(logging.getLogger('dliswriter'))

    pargs = make_parser().parse_args()

    prepare_directory(pargs.output_file_name, overwrite=pargs.overwrite)

    tmp_file_name = Path(pargs.output_file_name).resolve().parent/'_tmp.h5'
    pargs.input_file_name = tmp_file_name

    df = create_dlis_spec(
        n_points=int(pargs.n_points),
        n_images=pargs.n_images,
        n_cols=pargs.n_cols,
        time_based=pargs.time_based
    )

    df.write(
        pargs.output_file_name,
        input_chunk_size=int(pargs.input_chunk_size),
        output_chunk_size=int(pargs.output_chunk_size)
    )


if __name__ == '__main__':
    main()



from dlisio import dlis    # type: ignore  # untyped library


def test_file_header(short_dlis: dlis.file.LogicalFile) -> None:
    """Check attributes of DLIS file header."""

    header = short_dlis.fileheader
    assert header.id == "DEFAULT FHLR"
    assert header.sequencenr == "1"
    assert header.origin == 42



import pytest
from dlisio import dlis    # type: ignore  # untyped library

from tests.common import check_list_of_objects


def test_process(short_dlis: dlis.file.LogicalFile) -> None:
    """Check that the number of Process objects in the DLIS file matches the expected one."""

    procs = short_dlis.processes
    assert len(procs) == 2


@pytest.mark.parametrize(("idx", "name", "input_channels", "output_channels", "input_compts", "output_compts"), (
        (0, "Process 1", ["radius"], ["amplitude", "Channel 2"], ["COMPT-1"], ["COMPT2"]),
        (1, "Prc2", ["Channel 1"], ["Channel 2"], ["COMPT2", "COMPT-1"], []),
))
def test_process_params(short_dlis: dlis.file.LogicalFile, idx: int, name: str, input_channels: list[str],
                        output_channels: list[str], input_compts: list[str], output_compts: list[str]) -> None:
    """Check attributes of Process objects in the new DLIS file."""

    proc = short_dlis.processes[idx]

    assert proc.name == name
    assert proc.origin == 42

    check_list_of_objects(proc.input_channels, input_channels)
    check_list_of_objects(proc.output_channels, output_channels)
    check_list_of_objects(proc.input_computations, input_compts)
    check_list_of_objects(proc.output_computations, output_compts)



import pytest
import numpy as np
from pathlib import Path
import h5py  # type: ignore  # untyped library
import os
from typing import Generator

from dliswriter.logical_record import eflr_types
from dliswriter.utils.source_data_wrappers import NumpyDataWrapper

from tests.common import load_dlis
from tests.dlis_files_for_testing import write_short_dlis


@pytest.fixture(scope="session")
def base_data_path() -> Path:
    """Path to the resources files."""

    return Path(__file__).resolve().parent


@pytest.fixture(scope="session")
def reference_data_path(base_data_path: Path) -> Path:
    """Path to the reference HDF5 data file."""

    return base_data_path / "resources/mock_data.hdf5"


@pytest.fixture(scope="session")
def reference_data(reference_data_path: Path) -> Generator:
    """The reference HDF5 data file, open in read mode."""

    f = h5py.File(reference_data_path, "r")
    yield f
    f.close()


@pytest.fixture(scope="session")
def short_reference_data_path(base_data_path: Path) -> Path:
    """Path to the HDF5 file with short version of the reference data."""

    return base_data_path / "resources/mock_data_short.hdf5"


@pytest.fixture(scope="session")
def short_reference_data(short_reference_data_path: Path) -> Generator:
    """The reference short HDF5 data file, open in read mode."""

    f = h5py.File(short_reference_data_path, "r")
    yield f
    f.close()


@pytest.fixture(scope="session")
def short_dlis(short_reference_data_path: Path, base_data_path: Path) -> Generator:
    """A freshly written DLIS file - used in tests to check if all contents are there as expected."""

    dlis_path = base_data_path / "outputs/new_fake_dlis_shared.DLIS"
    write_short_dlis(dlis_path, data=short_reference_data_path)

    with load_dlis(dlis_path) as f:
        yield f

    if dlis_path.exists():  # does not exist if file creation failed
        os.remove(dlis_path)


@pytest.fixture
def new_dlis_path(base_data_path: Path) -> Generator:
    """Path for a new DLIS file to be created. The file is removed afterwards."""

    new_path = base_data_path / "outputs/new_fake_dlis.DLIS"
    os.makedirs(new_path.parent, exist_ok=True)
    yield new_path

    if new_path.exists():  # does not exist if file creation failed
        os.remove(new_path)


@pytest.fixture
def channel_parent() -> eflr_types.ChannelSet:
    return eflr_types.ChannelSet()


@pytest.fixture
def channel1(channel_parent: eflr_types.ChannelSet) -> eflr_types.ChannelItem:
    return eflr_types.ChannelItem("Channel 1", parent=channel_parent)


@pytest.fixture
def channel2(channel_parent: eflr_types.ChannelSet) -> eflr_types.ChannelItem:
    return eflr_types.ChannelItem("Channel 2", parent=channel_parent)


@pytest.fixture
def channel3(channel_parent: eflr_types.ChannelSet) -> eflr_types.ChannelItem:
    return eflr_types.ChannelItem("Channel 3", parent=channel_parent)


@pytest.fixture
def chan(channel_parent: eflr_types.ChannelSet) -> Generator:
    """Mock ChannelItem instance for tests."""

    yield eflr_types.ChannelItem("some_channel", parent=channel_parent)


@pytest.fixture
def channels(
    channel1: eflr_types.ChannelItem,
    channel2: eflr_types.ChannelItem,
    channel3: eflr_types.ChannelItem,
    chan: eflr_types.ChannelItem,
) -> dict:
    return {
        "Channel 1": channel1,
        "Channel 2": channel2,
        "Channel 3": channel3,
        "some_channel": chan,
    }


@pytest.fixture
def mock_data() -> NumpyDataWrapper:
    """Mock data (structured numpy array) for tests."""

    dt = np.dtype(
        [("time", float), ("amplitude", float, (10,)), ("radius", float, (12,))]
    )
    return NumpyDataWrapper(np.zeros(30, dtype=dt))


@pytest.fixture
def ccoef1() -> eflr_types.CalibrationCoefficientItem:
    return eflr_types.CalibrationCoefficientItem(
        "COEF-1", parent=eflr_types.CalibrationCoefficientSet()
    )


@pytest.fixture
def cmeasure1() -> eflr_types.CalibrationMeasurementItem:
    return eflr_types.CalibrationMeasurementItem(
        "CMEASURE-1", parent=eflr_types.CalibrationMeasurementSet()
    )


@pytest.fixture
def axis1() -> eflr_types.AxisItem:
    return eflr_types.AxisItem("Axis-1", parent=eflr_types.AxisSet())


@pytest.fixture
def param_parent() -> eflr_types.ParameterSet:
    return eflr_types.ParameterSet()


@pytest.fixture
def param1(param_parent: eflr_types.ParameterSet) -> eflr_types.ParameterItem:
    return eflr_types.ParameterItem("Param-1", parent=param_parent)


@pytest.fixture
def param2(param_parent: eflr_types.ParameterSet) -> eflr_types.ParameterItem:
    return eflr_types.ParameterItem("Param-2", parent=param_parent)


@pytest.fixture
def param3(param_parent: eflr_types.ParameterSet) -> eflr_types.ParameterItem:
    return eflr_types.ParameterItem("Param-3", parent=param_parent)


@pytest.fixture
def zone_parent() -> eflr_types.ZoneSet:
    return eflr_types.ZoneSet()


@pytest.fixture
def zone1(zone_parent: eflr_types.ZoneSet) -> eflr_types.ZoneItem:
    return eflr_types.ZoneItem("Zone-1", parent=zone_parent)


@pytest.fixture
def zone2(zone_parent: eflr_types.ZoneSet) -> eflr_types.ZoneItem:
    return eflr_types.ZoneItem("Zone-2", parent=zone_parent)


@pytest.fixture
def zone3(zone_parent: eflr_types.ZoneSet) -> eflr_types.ZoneItem:
    return eflr_types.ZoneItem("Zone-3", parent=zone_parent)


@pytest.fixture
def zones(
    zone1: eflr_types.ZoneItem, zone2: eflr_types.ZoneItem, zone3: eflr_types.ZoneItem
) -> dict:
    return {"Zone-1": zone1, "Zone-2": zone2, "Zone-3": zone3}


@pytest.fixture
def process_parent() -> eflr_types.ProcessSet:
    return eflr_types.ProcessSet()


@pytest.fixture
def process1(process_parent: eflr_types.ProcessSet) -> eflr_types.ProcessItem:
    return eflr_types.ProcessItem("Process 1", parent=process_parent)


@pytest.fixture
def process2(process_parent: eflr_types.ProcessSet) -> eflr_types.ProcessItem:
    return eflr_types.ProcessItem("Prc2", parent=process_parent)


@pytest.fixture
def channel_group(
    channel1: eflr_types.ChannelItem,
    channel2: eflr_types.ChannelItem,
    channel3: eflr_types.ChannelItem,
) -> eflr_types.GroupItem:
    return eflr_types.GroupItem(
        "Group of channels",
        object_type="CHANNEL",
        object_list=[channel1, channel2, channel3],
        parent=eflr_types.GroupSet(),
    )


@pytest.fixture
def process_group(
    process2: eflr_types.ProcessItem, process1: eflr_types.ProcessItem
) -> eflr_types.GroupItem:
    return eflr_types.GroupItem(
        "Group of processes",
        object_type="PROCESS",
        object_list=[process1, process2],
        parent=eflr_types.GroupSet(),
    )


@pytest.fixture
def computation1() -> eflr_types.ComputationItem:
    return eflr_types.ComputationItem("Compt1", parent=eflr_types.ComputationSet())


@pytest.fixture
def computation2() -> eflr_types.ComputationItem:
    return eflr_types.ComputationItem("CMPT-2", parent=eflr_types.ComputationSet())


@pytest.fixture()
def frame(
    channel1: eflr_types.ChannelItem,
    channel2: eflr_types.ChannelItem,
    channel3: eflr_types.ChannelItem,
) -> eflr_types.FrameItem:
    return eflr_types.FrameItem(
        "MAIN FRAME",
        channels=(channel1, channel2, channel3),
        index_type="BOREHOLE-DEPTH",
        parent=eflr_types.FrameSet(),
    )


@pytest.fixture()
def well_reference_point() -> eflr_types.WellReferencePointItem:
    return eflr_types.WellReferencePointItem(
        name="AQLN WELL-REF",
        permanent_datum="AQLN permanent_datum",
        vertical_zero="AQLN vertical_zero",
        permanent_datum_elevation=1234.51,
        above_permanent_datum=888.51,
        magnetic_declination=999.51,
        coordinate_1_name="Latitude",
        coordinate_1_value=40.395240,
        coordinate_2_name="Longitude",
        coordinate_2_value=27.792470,
        parent=eflr_types.WellReferencePointSet(),
    )
